/*----------------------------------------------------------------------------*/
/*test.c*/
/*----------------------------------------------------------------------------*/
/*A Python C extension module that allows programs in Python to retrieve
  the uptime and load averages.*/
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
#include <Python.h>
/*----------------------------------------------------------------------------*/
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
/*----------------------------------------------------------------------------*/
#include <hurd.h>
#include <ps.h>
#include <time.h>
#include <sys/time.h>
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/*Retrieves the uptime of the system*/
static
PyObject *
utld_uptime
	(
	PyObject * self,	/*owner object*/
	PyObject * args		/*set of arguments*/
 	);
/*----------------------------------------------------------------------------*/
/*Retrieves the load averages*/
static
PyObject *
utld_loadavgs
	(
	PyObject * self,	/*owner object*/
	PyObject * args		/*set of arguments*/
	);
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/*The array of descriptions of methods*/
static PyMethodDef utldMethods[] =
	{
	{"uptime",   utld_uptime,   METH_VARARGS, "Retrieve the uptime of the system"},
	{"loadavgs", utld_loadavgs, METH_VARARGS, "Retrieve the load averages"},
	{NULL, NULL, 0, NULL}
	};
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/*--------Global Functions----------------------------------------------------*/
/*Module initialization function*/
PyMODINIT_FUNC
initutld(void)
	{
	/*Perform the initialization*/
	Py_InitModule("utld", utldMethods);
	}/*initutld*/;
/*----------------------------------------------------------------------------*/
/*Retrieves the uptime of the system*/
static
PyObject *
utld_uptime
	(
	PyObject * self,	/*owner object*/
	PyObject * args		/*set of arguments*/
	)
	{
	/*The result of operations*/
	error_t err;

	/*Obtais the time of creation of init*/
	int
	get_boot_time
		(
		struct timeval * tm /*the receiver of the result*/
		)
		{
		/*The new context*/
		struct ps_context * context;

		/*The holder of statistical information about init*/
		struct proc_stat * stat;

		/*Try to create a new context for the current process*/
		err = ps_context_create(getproc(), &context);
		if(err)
			return -1;
			
		/*Try to find the statistical information about init*/
		err = ps_context_find_proc_stat(context, 1/*init's pid*/, &stat);

		if(err)
			{
			ps_context_free(context);
			return -2;
			}
			
		/*Require only the basic information*/
		err = proc_stat_set_flags(stat, PSTAT_TASK_BASIC);
		
		/*Set an error, even if the previous operation has not failed,
		 but the flag has not been set*/
		if(!err && !(stat->flags & PSTAT_TASK_BASIC))
			err = EGRATUITOUS;
			
		/*Stop if there are errors*/
		if(err)
			{
			ps_context_free(context);
			return -3;
			}
			
		/*Obtain the pointer to the time of creation of init*/
		struct time_value * const tv = &proc_stat_task_basic_info(stat)->creation_time;
		
		/*Copy the fields of the result into the parameter*/
		tm->tv_sec  = tv->seconds;
		tm->tv_usec = tv->microseconds;
		
		/*Free the process context*/
		ps_context_free(context);
		
		/*Return success*/
		return 0;
		}/*get_boot_time*/

	/*Obtain the time of boot*/
	struct timeval boot_time;
	err = get_boot_time(&boot_time);

	/*If an error has occurred*/
	if(err)
		/*return error number*/
		return Py_BuildValue("i", err);
	
	/*Obtain the current time*/
	struct timeval current_time;
	err = gettimeofday(&current_time, NULL);

	/*If an error has occurred*/
	if(err)
		/*return None*/
		return Py_BuildValue("");

	/*Obtain the uptime*/
	struct timeval uptime;
	timersub(&current_time, &boot_time, &uptime);

	/*Return the information as a tuple (seconds, microseconds)*/
	return Py_BuildValue("(ll)", (long)uptime.tv_sec, (long)uptime.tv_usec);
	}/*ultd_uptime*/
/*----------------------------------------------------------------------------*/
/*Retrieves the load averages*/
static
PyObject *
utld_loadavgs
	(
	PyObject * self,	/*owner object*/
	PyObject * args		/*set of arguments*/
	)
	{
	/*The result of operations*/
	error_t err;
	
	/*The new context*/
	struct ps_context * context;

	/*The list of proc_stats*/
	struct proc_stat_list * procs;

	/*Try to create a new context for the current process*/
	err = ps_context_create(getproc(), &context);
	if(err)
		return Py_BuildValue("s", "ps_context_create");
		
	/*Try to create the list of items statistical information for the current context*/
	err = proc_stat_list_create(context, &procs);
	if(err)
		{
		ps_context_free(context);
		return Py_BuildValue("s", "proc_stat_list_create");
		}
		
	/*Try to obtain the load averages*/
	struct host_load_info * load;
	err = ps_host_load_info(&load);
	if(err)
		{
		ps_context_free(context);
		return Py_BuildValue("s", "ps_host_load_info");
		}
	
	/*Free the process context*/
	ps_context_free(context);
	
	/*Return the three-tuple containing the properly scaled load averages*/
	return Py_BuildValue
		(
		"(ddd)",
		load->avenrun[0] / (float)LOAD_SCALE,
		load->avenrun[1] / (float)LOAD_SCALE,
		load->avenrun[2] / (float)LOAD_SCALE
		);
	}/*utld_loadavgs*/;
/*----------------------------------------------------------------------------*/
